/* signals.c -- a file which should cantain some *REAL* signals management. */

/* Copyright (C) 1993, 1994, 1995 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* Written by Tudor Hulubei and Andrei Pitis.  */


#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#else /* !HAVE_STDLIB_H */
#include "ansi_stdlib.h"
#endif /* !HAVE_STDLIB_H */

#include <signal.h>

#include "signals.h"
#include "inputline.h"
#include "tty.h"
#include "misc.h"


extern int  pid;
extern char *screen;
extern input_line *input;


int suspend_allowed = OFF;
int signals_status;
int UserHeartAttack = 0;


RETSIGTYPE suspend(signum)
    int signum;
{
    signal(SIGTSTP, suspend);

    if (suspend_allowed)
    {
	tty_set_mode(TTY_CANONIC);
	tty_put_screen(screen);
	kill(pid, SIGSTOP);
    }
}


RETSIGTYPE resume(signum)
    int signum;
{
    extern int current_mode;
    extern void refresh_after_suspend __P((int));

    signal(SIGCONT, resume);

    refresh_after_suspend(current_mode);
}


RETSIGTYPE user_panic(signum)
    int signum;
{
    signal(signum, user_panic);

    UserHeartAttack = 1; 
}


RETSIGTYPE fatal_signal(signum)
    int signum;
{
    extern void clean_up __P(());

    clean_up();

    switch (signum)
    {
	case SIGTERM:
	case SIGQUIT:

	    fprintf(stderr, "%s: got %s.\n",
	           program, ((signum == SIGTERM) ? "SIGTERM" : "SIGQUIT"));
	    break;

	case SIGSEGV:

	    fprintf(stderr, "%s: got SIGSEGV.\n", program);
	    goto ask_report;
	    break;

	case SIGHUP:

	    /* :-) */
	    fprintf(stderr, "%s: got SIGHUP.\n", program);
	    break;

	default:

	    fprintf(stderr,
	    	    "%s: got a stupid signal (%d). Unless it was a joke ...\n",
	    	    program, signum);
	  ask_report:
	    fprintf(stderr, "%s: please report to tudor@cs.pub.ro.\n",
	    	    program);
	    break;
    }

    exit(1);
}


void signals_dfl()
{
    signal(SIGTERM, SIG_DFL);
    signal(SIGQUIT, SIG_DFL);
    signal(SIGINT , SIG_DFL);
}


void signals(status)
    int status;
{
    signals_status = status;
    signal(SIGTERM, status == ON ? fatal_signal : SIG_IGN);
    signal(SIGQUIT, status == ON ? fatal_signal : SIG_IGN);
    signal(SIGINT , status == ON ? user_panic   : SIG_IGN);
}


void ignore_signals()
{
    signal(SIGILL,  SIG_IGN);
    signal(SIGTRAP, SIG_IGN);
    signal(SIGABRT, SIG_IGN);
    signal(SIGUSR1, SIG_IGN);
    signal(SIGUSR2, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    signal(SIGFPE,  SIG_IGN);
}


void restore_signals()
{
    signal(SIGILL,  SIG_DFL);
    signal(SIGTRAP, SIG_DFL);
    signal(SIGABRT, SIG_DFL);
    signal(SIGUSR1, SIG_DFL);
    signal(SIGUSR2, SIG_DFL);
    signal(SIGPIPE, SIG_DFL);
    signal(SIGFPE,  SIG_DFL);
}
